import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pathlib

matplotlib.style.use(pathlib.Path(__file__).parent / pathlib.Path("piel_fast.rcParams"))

__all__ = [
    "plot_simple",
    "plot_simple_multi_row",
]


def plot_simple(x_data: np.array, y_data: np.array, ylabel: str, xlabel: str):
    """
    Plot a simple line graph.

    Args:
        x_data (np.array): X axis data.
        y_data (np.array): Y axis data.
        ylabel (str): Y axis label.
        xlabel (str): X axis label.

    Returns:
        plt: Matplotlib plot.
    """
    plt.plot(x_data, y_data)
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    return plt


def plot_simple_multi_row(
    data: pd.DataFrame,
    x_axis_column_name: str = "t",
    row_list: list | None = None,
    y_axis_title_list: list | None = None,
    x_axis_title: str | None = None,
):
    """
    Plot multiple rows of data on the same plot. Each row is a different line. Each row is a different y axis. The x
    axis is the same for all rows. The y axis title is the same for all rows.

    Args:
        data (pd.DataFrame): Data to plot.
        x_axis_column_name (str, optional): Column name of the x axis. Defaults to "t".
        row_list (list, optional): List of column names to plot. Defaults to None.
        y_axis_title_list (list, optional): List of y axis titles. Defaults to None.
        x_axis_title (str, optional): Title of the x axis. Defaults to None.

    Returns:
        plt: Matplotlib plot.
    """
    x = data[x_axis_column_name]
    y_array = []

    if y_axis_title_list is None:
        y_axis_title_list = row_list

    row_amount = len(row_list)
    for row_name in row_list:
        y_array.append(data[row_name])

    fig, axes = plt.subplots(row_amount, 1, sharex=True)

    for i in range(len(row_list)):
        axes[i].plot(x, y_array[i])
        axes[i].grid(True)
        axes[i].set(ylabel=y_axis_title_list[i])

    # TODO Xaxis title
    # TODO align all ytitles

    return plt



"""Top-level package for piel."""
import os
import pathlib

# Libraries
from piel import models  # NOQA: F401
from piel import visual  # NOQA: F401

# Functions

from .config import *
from .defaults import *
from .file_system import *
from .integration import *
from .parametric import *
from .tools import *


os.environ["PIEL_PACKAGE_DIRECTORY"] = str(
    pathlib.Path(__file__).parent.parent.resolve()
)

__author__ = """Dario Quintero"""
__email__ = "darioaquintero@gmail.com"
__version__ = "0.0.47"



import openlane
from piel.config import piel_path_types
from piel.file_system import return_path, read_json
from piel.defaults import test_spm_open_lane_configuration

__all__ = ["run_openlane_flow"]


def run_openlane_flow(
    configuration: dict | None = test_spm_open_lane_configuration,
    design_directory: piel_path_types = "/foss/designs/spm",
) -> None:
    """
    Runs the OpenLane flow.

    Args:
        configuration(dict): OpenLane configuration dictionary. If none is present it will default to the config.json file on the design_directory.
        design_directory(piel_path_types): Design directory PATH.

    Returns:
        None
    """
    design_directory = return_path(design_directory)
    if configuration is None:
        # Get extract configuration file from config.json on directory
        config_json_filepath = design_directory / "config.json"
        configuration = read_json(config_json_filepath)

    Classic = openlane.Flow.get("Classic")

    flow = Classic(
        config=configuration,
        design_dir=str(design_directory.resolve()),
    )

    flow.start()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "bokeh",
    "cython==0.29.21",
    "jupyter_bokeh",
    "jupytext",
    "Click>=7.0",
    "cocotb",
    "femwell",
    "hdl21>=4",
    "jax",
    "jaxlib",
    "gdsfactory>=7",
    "networkx",
    "openlane",
    "pandas",
    "sax>=0.8.8",
    "thewalrus",
    "vlsir>=4",
    "qutip",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Dario Quintero",
    author_email="darioaquintero@gmail.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Photonic Integrated Electronics: microservices to codesign photonics, electronics, communications, quantum, and more.",
    entry_points={
        "console_scripts": [
            "piel=piel.cli:main",
        ],
    },
    extras_require={
        "develop": [
            "sphinx",
            "sphinx_autodoc_typehints",
            "sphinx-pydantic",
            "sphinx-autoapi",
            "sphinx-autobuild",
            "sphinx_rtd_theme",
            "sphinx-gallery",
            "nbsphinx",
            "myst_parser",
            "pandoc",
            "flake8",
        ]
    },
    install_requires=requirements,
    license="MIT license",
    long_description_content_type="text/markdown",
    long_description=readme,
    include_package_data=True,
    keywords="piel",
    name="piel",
    packages=find_packages(include=["piel", "piel.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/daquintero/piel",
    version="0.0.47",
    zip_safe=False,
)



"""
There are a number of ways to generate gdsfactory integration.

It is worth noting that GDSFactory has already the following PDKs installed:
* SKY130nm https://gdsfactory.github.io/skywater130/
* GF180nm https://gdsfactory.github.io/gf180/
"""
import gdsfactory as gf
from ..config import piel_path_types
from ..file_system import check_path_exists
from piel.tools.openlane.migrate import get_design_from_openlane_migration
from piel.tools.openlane import find_design_run

__all__ = ["create_gdsfactory_component_from_openlane"]


def create_gdsfactory_component_from_openlane(
    design_name_v1: str | None = None,
    design_directory: piel_path_types | None = None,
    run_name: str | None = None,
    v1: bool = True,
) -> gf.Component:
    """
    This function cretes a gdsfactory layout component that can be included in the network codesign of the device, or that can be used for interconnection codesign.

    It will look into the latest design run and extract the final OpenLane-generated GDS. You do not have to have run this with OpenLane2 as it just looks at the latest run.

    Args:
        design_name_v1(str): Design name of the v1 design that can be found within `$OPENLANE_ROOT/"<latest>"/designs`.
        design_directory(piel_path_types): Design directory PATH.
        run_name(str): Name of the run to extract the GDS from. If None, it will look at the latest run.
        v1(bool): If True, it will import the design from the OpenLane v1 configuration.

    Returns:
        component(gf.Component): GDSFactory component.
    """
    design_name, design_directory = get_design_from_openlane_migration(
        v1=v1, design_name_v1=design_name_v1, design_directory=design_directory
    )
    latest_design_run_directory = find_design_run(design_directory, run_name=run_name)
    final_gds_run = (
        latest_design_run_directory
        / "results"
        / "final"
        / "gds"
        / (design_name + ".gds")
    )
    check_path_exists(final_gds_run, raise_errors=True)
    component = gf.import_gds(final_gds_run, name=design_name)
    return component



import jax.numpy as jnp
import qutip

from ..tools.thewalrus import unitary_permanent
from ..tools.qutip import (
    fock_state_nonzero_indexes,
    fock_state_to_photon_number_factorial,
    subunitary_selection_on_index,
)

__all__ = ["fock_transition_probability_amplitude"]


def fock_transition_probability_amplitude(
    initial_fock_state: qutip.Qobj,
    final_fock_state: qutip.Qobj,
    unitary_matrix: jnp.ndarray,
):
    """
    This function returns the transition probability amplitude between two Fock states when propagating in between
    the unitary_matrix which represents a quantum state circuit.

    Note that based on (TODO cite Jeremy), the initial Fock state corresponds to the columns of the unitary and the
    final Fock states corresponds to the rows of the unitary.

    .. math ::
        \newcommand{\ket}[1]{\left|{#1}\right\rangle}

    The subunitary :math:`U_{f_1}^{f_2}` is composed from the larger unitary by selecting the rows from the output state
    Fock state occupation of :math:`\ket{f_2}`, and columns from the input :math:`\ket{f_1}`. In our case, we need to select the
    columns indexes :math:`(0,3)` and rows indexes :math:`(1,2)`.

    If we consider a photon number of more than one for the transition Fock states, then the Permanent needs to be
    normalised. The probability amplitude for the transition is described as:

    .. math ::
        a(\ket{f_1} \to \ket{f_2}) = \frac{\text{per}(U_{f_1}^{f_2})}{\sqrt{(j_1! j_2! ... j_N!)(j_1^{'}! j_2^{'}! ... j_N^{'}!)}}

    Args:
        initial_fock_state (qutip.Qobj): A QuTip QObj representation of the initial Fock state.
        final_fock_state (qutip.Qobj): A QuTip QObj representation of the final Fock state.
        unitary_matrix (jnp.ndarray): A JAX NumPy array representation of the unitary matrix.

    Returns:
        float: The transition probability amplitude between the initial and final Fock states.
    """
    columns_indices = fock_state_nonzero_indexes(initial_fock_state)
    rows_indices = fock_state_nonzero_indexes(final_fock_state)

    initial_fock_state_photon_number_factorial = fock_state_to_photon_number_factorial(
        initial_fock_state
    )
    final_fock_state_photon_number_factorial = fock_state_to_photon_number_factorial(
        final_fock_state
    )

    subunitary_selection = subunitary_selection_on_index(
        unitary_matrix=unitary_matrix,
        rows_index=rows_indices,
        columns_index=columns_indices,
    )

    transition_probability_amplitude = unitary_permanent(subunitary_selection)[0] / (
        jnp.sqrt(
            initial_fock_state_photon_number_factorial
            * final_fock_state_photon_number_factorial
        )
    )
    return transition_probability_amplitude



"""
These functions provide easy tools for easily migrating between OpenLane v1 and v2 based designs.
"""
import pathlib
from .v1 import get_design_directory_from_root_openlane_v1
from piel.file_system import return_path

__all__ = [
    "get_design_from_openlane_migration",
]


def get_design_from_openlane_migration(
    v1: bool = True,
    design_name_v1: str | None = None,
    design_directory: str | pathlib.Path | None = None,
    root_directory_v1: str | pathlib.Path | None = None,
) -> (str, pathlib.Path):
    """
    This function provides the integration mechanism for easily migrating the interconnection with other toolsets from an OpenLane v1 design to an OpenLane v2 design.

    This function checks if the inputs are to be treated as v1 inputs. If so, and a `design_name` is provided then it will set the `design_directory` to the corresponding `design_name` directory in the corresponding `root_directory_v1 / designs`. If no `root_directory` is provided then it returns `$OPENLANE_ROOT/"<latest>"/. If a `design_directory` is provided then this will always take precedence even with a `v1` flag.

    Args:
        v1(bool): If True, it will migrate from v1 to v2.
        design_name_v1(str): Design name of the v1 design that can be found within `$OPENLANE_ROOT/"<latest>"/designs`.
        design_directory(str): Design directory PATH. Optional path for v2-based designs.
        root_directory_v1(str): Root directory of OpenLane v1. If set to None it will return `$OPENLANE_ROOT/"<latest>"`

    Returns:
        None
    """
    if design_directory is not None:
        design_directory = return_path(design_directory)
        design_name = design_directory.name
        return design_name, design_directory
    elif v1:
        design_directory = get_design_directory_from_root_openlane_v1(
            design_name=design_name_v1, root_directory=root_directory_v1
        )
        design_name = design_name_v1
        return design_name, design_directory
    else:
        raise ValueError(
            "You must provide either a design_directory or a design_name_v1"
        )



import pandas as pd
from .file_system import return_path
from .config import piel_path_types

__all__ = [
    "read_csv_to_pandas",
    "read_vcd_to_json",
]


def read_csv_to_pandas(file_path: piel_path_types):
    """
    This function returns a Pandas dataframe that contains all the simulation data outputted from the simulation run.
    """
    file_path = return_path(file_path)
    simulation_data = pd.read_csv(file_path)
    return simulation_data


def read_vcd_to_json(file_path: piel_path_types):
    from pyDigitalWaveTools.vcd.parser import VcdParser

    file_path = return_path(file_path)
    with open(str(file_path.resolve())) as vcd_file:
        vcd = VcdParser()
        vcd.parse(vcd_file)
        json_data = vcd.scope.toJson()
    return json_data



"""
This file provides a set of utilities that allow much easier integration between `sax` and the relevant tools that we use.
"""
import jax.numpy as jnp
import sax
from ..gdsfactory.netlist import get_matched_ports_tuple_index
from typing import Optional  # NOQA : F401

__all__ = ["get_sdense_ports_index", "sax_to_s_parameters_standard_matrix", "snet"]


def get_sdense_ports_index(input_ports_order: tuple, all_ports_index: dict) -> dict:
    """
    This function returns the ports index of the sax dense S-parameter matrix.

    Given that the order of the iteration is provided by the user, the dictionary keys will also be ordered
    accordingly when iterating over them. This requires the user to provide a set of ordered.

    TODO verify reasonable iteration order.

    .. code-block:: python

        # The input_ports_order can be a tuple of tuples that contain the index and port name. Eg.
        input_ports_order = ((0, "in_o_0"), (5, "in_o_1"), (6, "in_o_2"), (7, "in_o_3"))
        # The all_ports_index is a dictionary of the ports index. Eg.
        all_ports_index = {
            "in_o_0": 0,
            "out_o_0": 1,
            "out_o_1": 2,
            "out_o_2": 3,
            "out_o_3": 4,
            "in_o_1": 5,
            "in_o_2": 6,
            "in_o_3": 7,
        }
        # Output
        {"in_o_0": 0, "in_o_1": 5, "in_o_2": 6, "in_o_3": 7}

    Args:
        input_ports_order (tuple): The ports order tuple. Can be a tuple of tuples that contain the index and port name.
        all_ports_index (dict): The ports index dictionary.

    Returns:
        tuple: The ordered input ports index tuple.
    """
    # TODO look into jnp.at method https://github.com/flaport/sax/blob/a87c3bf8c792dc227779e5d010627897f4cd8278/sax/typing_.py#L355
    input_ports_index = {key: all_ports_index[key] for key in input_ports_order}
    return input_ports_index


def sax_to_s_parameters_standard_matrix(
    sax_input: sax.SType,
    input_ports_order: tuple | None = None,
) -> tuple:
    """
    A ``sax`` S-parameter SDict is provided as a dictionary of tuples with (port0, port1) as the key. This
    determines the direction of the scattering relationship. It means that the number of terms in an S-parameter
    matrix is the number of ports squared.

    In order to generalise, this function returns both the S-parameter matrices and the indexing ports based on the
    amount provided. In terms of computational speed, we definitely would like this function to be algorithmically
    very fast. For now, I will write a simple python implementation and optimise in the future.

    It is possible to see the `sax` SDense notation equivalence here:
    https://flaport.github.io/sax/nbs/08_backends.html

    .. code-block:: python

        import jax.numpy as jnp
        from sax.core import SDense

        # Directional coupler SDense representation
        dc_sdense: SDense = (
            jnp.array([[0, 0, τ, κ], [0, 0, κ, τ], [τ, κ, 0, 0], [κ, τ, 0, 0]]),
            {"in0": 0, "in1": 1, "out0": 2, "out1": 3},
        )


        # Directional coupler SDict representation
        # Taken from https://flaport.github.io/sax/nbs/05_models.html
        def coupler(*, coupling: float = 0.5) -> SDict:
            kappa = coupling**0.5
            tau = (1 - coupling) ** 0.5
            sdict = reciprocal(
                {
                    ("in0", "out0"): tau,
                    ("in0", "out1"): 1j * kappa,
                    ("in1", "out0"): 1j * kappa,
                    ("in1", "out1"): tau,
                }
            )
            return sdict

    If we were to relate the mapping accordingly based on the ports indexes, a S-Parameter matrix in the form of
    :math:`S_{(output,i),(input,i)}` would be:

    .. math::

        S = \\begin{bmatrix}
                S_{00} & S_{10} \\\\
                S_{01} & S_{11} \\\\
            \\end{bmatrix} =
            \\begin{bmatrix}
            \\tau & j \\kappa \\\\
            j \\kappa & \\tau \\\\
            \\end{bmatrix}

    Note that the standard S-parameter and hence unitary representation is in the form of:

    .. math::

        S = \\begin{bmatrix}
                S_{00} & S_{01} \\\\
                S_{10} & S_{11} \\\\
            \\end{bmatrix}


    .. math::

        \\begin{bmatrix}
            b_{1} \\\\
            \\vdots \\\\
            b_{n}
        \\end{bmatrix}
        =
        \\begin{bmatrix}
            S_{11} & \\dots & S_{1n} \\\\
            \\vdots & \\ddots & \\vdots \\\\
            S_{n1} & \\dots & S_{nn}
        \\end{bmatrix}
        \\begin{bmatrix}
            a_{1} \\\\
            \\vdots \\\\
            a_{n}
        \\end{bmatrix}

    TODO check with Floris, does this mean we need to transpose the matrix?

    Args:
        sax_input (sax.SType): The sax S-parameter dictionary.
        input_ports_order (tuple): The ports order tuple containing the names and order of the input ports.

    Returns:
        tuple: The S-parameter matrix and the input ports index tuple in the standard S-parameter notation.
    """
    dense_s_parameter_matrix, dense_s_parameter_index = sax.sdense(sax_input)
    # print(dense_s_parameter_index)
    all_ports_list = dense_s_parameter_index.keys()
    # Now we get the indexes of the input ports that we care about to restructure the dense matrix with the columns
    # we care about.
    if input_ports_order is not None:
        output_ports_order = tuple(set(all_ports_list) - set(input_ports_order))
        (
            input_ports_index_tuple_order,
            input_matched_ports_name_tuple_order,
        ) = get_matched_ports_tuple_index(
            ports_index=dense_s_parameter_index,
            selected_ports_tuple=input_ports_order,
            sorting_algorithm="selected_ports",
        )
        (
            output_ports_index_tuple_order,
            output_matched_ports_name_tuple_order,
        ) = get_matched_ports_tuple_index(
            ports_index=dense_s_parameter_index,
            selected_ports_tuple=output_ports_order,
            sorting_algorithm="selected_ports",
        )
    else:
        (
            input_ports_index_tuple_order,
            input_matched_ports_name_tuple_order,
        ) = get_matched_ports_tuple_index(
            ports_index=dense_s_parameter_index, prefix="in"
        )
        (
            output_ports_index_tuple_order,
            output_matched_ports_name_tuple_order,
        ) = get_matched_ports_tuple_index(
            ports_index=dense_s_parameter_index, prefix="out"
        )

    output_ports_index_tuple_order_jax = jnp.asarray(output_ports_index_tuple_order)
    input_ports_index_tuple_order_jax = jnp.asarray(input_ports_index_tuple_order)
    # We now select the SDense columns that we care about.
    s_parameters_standard_matrix = dense_s_parameter_matrix.at[
        output_ports_index_tuple_order_jax
    ].get()
    s_parameters_standard_matrix = s_parameters_standard_matrix.at[
        :, input_ports_index_tuple_order_jax
    ].get()
    # Now we select the SDense rows that we care about after transposing the matrix.
    # TODO verify matrix transpose for unitary match. I think it is right.
    return s_parameters_standard_matrix.T, input_matched_ports_name_tuple_order


snet = sax_to_s_parameters_standard_matrix



import jax.numpy as jnp

__all__ = ["calculate_cross_sectional_area_m2"]


def calculate_cross_sectional_area_m2(
    diameter_m: float,
) -> float:
    """
    Calculates the cross sectional area of a circle in meters squared.

    Args:
        diameter_m (float): Diameter of the circle in meters.

    Returns:
        float: Cross sectional area in meters squared.
    """
    return jnp.pi * (diameter_m**2) / 4
